{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE GADTs #-}

--
-- Groebner Bases for polynomial ideals
--

module Ideal (
	Ideal(Ideal), ideal, basis,  -- data type, constructor, destructor
	contains, subset,            -- operations
	buchberger, minimize, reduce, spoly
) where

import Data.Function
import Data.List

import Monomial
import Polynomial


------------------------------------
-- Data type
------------------------------------

-- | An ideal is defined by its reduced groebner base
data Ideal t where
	Ideal :: (Eq t, Fractional t) => { basis :: [Polynomial t] } -> Ideal t

-- | Constructor constructs a sorted reduced Groebner basis
ideal :: (Eq t, Fractional t, Ord Monomial) => [Polynomial t] -> Ideal t
ideal = Ideal
	. reverse
	. sortBy (compare `on` lm)
	. reduce
	. map normalize
	. minimize
	. buchberger
	. filter (/= 0)

-- | Buchberger's Algorithm to find a Groebner Basis for the Ideal
buchberger :: (Eq t, Fractional t, Ord Monomial) => [Polynomial t] -> [Polynomial t]
buchberger ps
	| null ss   = ps
	| otherwise = buchberger (head ss : ps)
	where ss = [s | p <- ps, q <- ps, p /= q,
	                s <- [snd $ division (spoly p q) ps], s /= 0]

-- | Minimize a Groebner basis
minimize :: (Eq t, Fractional t, Ord Monomial) => [Polynomial t] -> [Polynomial t]
minimize ps
	| null ps'  = ps
	| otherwise = minimize (delete (head ps') ps)
	where
		ps' = [p | p <- ps, Ideal (map lt' $ delete p ps) `contains` lt' p]
		lt' = normalize . lt
-- see also Lemma 2.7.3 p. 89

-- | Produces a reduces Groebner basis from a minimized one
reduce :: (Eq t, Fractional t, Ord Monomial) => [Polynomial t] -> [Polynomial t]
reduce ps
	| null l    = ps
	| otherwise = reduce (p' : ps')
	where
		l       = [(p, as) | p <- ps, (_, m) <- coeff p,
					 (as, r) <- [division (fromMono m) (delete p ps)], r == 0]
		(p, as) = head l
		ps'     = delete p ps
		p'      = p - sum (zipWith (*) as ps')

-- | Construct the S-polynomial
spoly :: (Eq t, Fractional t, Ord Monomial) =>
			Polynomial t -> Polynomial t -> Polynomial t
spoly p q = cnst (lc q) * fromMono (s `divm` m) * p
					- cnst (lc p) * fromMono (s `divm` n) * q
	where
		m = lm p
		n = lm q
		s = lcmm m n


------------------------------------
-- Class instances
------------------------------------

instance Show t => Show (Ideal t) where
	show (Ideal ps) = "<< " ++ intercalate ", " (map show ps) ++ " >>"

instance (Eq t, Fractional t) => Eq (Ideal t) where
	(Ideal ps) == (Ideal qs) = ps == qs -- sorted by degree
-- The reduced Groebner base is uniquely determined, so equality of ideals boils
-- down to checking the set-equality of the Groebner bases.

-- I + J = { i + j | i in I and j in J }
-- I * J = { sum_k (ik * jk) | ik in I and jk in J for all k }
instance (Eq t, Fractional t, Ord Monomial) => Num (Ideal t) where
	(Ideal ps) + (Ideal qs) = ideal (ps ++ qs)
	(Ideal ps) * (Ideal qs) = ideal [p * q | p <- ps, q <- qs]
	negate                  = id
	abs _                   = undefined
	signum _                = undefined
	fromInteger 0           = Ideal []  -- {0} as an ideal
	fromInteger _           = Ideal [cnst 1] -- <1> as an ideal


------------------------------------
-- Operations
------------------------------------

-- | Membership test for ideals
contains :: (Eq t, Fractional t, Ord Monomial) => Ideal t -> Polynomial t -> Bool
contains (Ideal ps) q = snd (division q ps) == 0

-- | Is an ideal contained in another?
subset :: (Eq t, Fractional t, Ord Monomial) => Ideal t -> Ideal t -> Bool
subset (Ideal ps) i = all (i `contains`) ps

